theory Labs05Part1

imports Main

begin

datatype (set: 'a) list = nil | cons 'a "'a list"

declare [[names_short]]

fun app :: "'a list \<Rightarrow> 'a list \<Rightarrow> 'a list" where
  "app nil ys = ys"
| "app (cons x xs) ys = cons x (app xs ys)"

fun reverse :: "'a list \<Rightarrow> 'a list" where
  "reverse nil = nil"
| "reverse (cons x xs) = app (reverse xs) (cons x nil)"

lemma app_nil_manual: "app xs nil = xs"
  proof (induction xs)
    case nil
    then show ?case by auto
  next
    case (cons x1 xs)
    have 1: "app (cons x1 xs) nil = cons x1 (app xs nil)" by simp
    have 2: "cons x1 (app xs nil) = cons x1 xs" by (simp add: cons.IH)
    show "app (cons x1 xs) nil = cons x1 xs" using 1 2 by simp
  qed

lemma app_nil [simp]: "app xs nil = xs"
  apply (induction xs)
  by auto

lemma app_assoc: "app xs (app ys zs) = app (app xs ys) zs"
  sorry

lemma reverse_app [simp]: "reverse (app xs ys) = app (reverse ys) (reverse xs)"
  apply (induction xs)
  by (auto simp add: app_assoc)

lemma reverse_reverse [simp]: "reverse (reverse xs) = xs"
  apply (induction xs)
  by auto

lemma rev_injective: "reverse xs1 = reverse xs2 \<Longrightarrow> xs1 = xs2"
  sorry

fun fold_left :: "'a list \<Rightarrow> 'b \<Rightarrow> ('b \<Rightarrow> 'a \<Rightarrow> 'b) \<Rightarrow> 'b" where
  "fold_left nil acc f = acc"
| "fold_left (cons x xs) acc f = fold_left xs (f acc x) f"

fun fold_right :: "'a list \<Rightarrow> 'b \<Rightarrow> ('a \<Rightarrow> 'b \<Rightarrow> 'b) \<Rightarrow> 'b" where
  "fold_right nil b f = b"
| "fold_right (cons x xs) b f = f x (fold_right xs b f)"

(* Write and prove the lemmas about `fold_left` and `fold_right` here *)

datatype (set: 'a) tree = Leaf | Node "'a tree" 'a "'a tree"

fun tree_to_list :: "'a tree \<Rightarrow> 'a list" where
  "tree_to_list Leaf = nil"
| "tree_to_list (Node l v r) = app (tree_to_list l) (cons v (tree_to_list r))"

fun tree_insert :: "int tree \<Rightarrow> int \<Rightarrow> int tree" where
  "tree_insert Leaf v = Node Leaf v Leaf"
| "tree_insert (Node l v' r) v = (
    if (v < v')
    then Node (tree_insert l v) v' r
    else Node l v' (tree_insert r v)
  )"

fun is_search_tree :: "int tree \<Rightarrow> bool" where
  "is_search_tree Leaf = True"
| "is_search_tree (Node l v r) = (
    is_search_tree l \<and>
    is_search_tree r \<and>
    (\<forall> v' \<in> tree.set l. v' < v) \<and>
    (\<forall> v' \<in> tree.set r. v' \<ge> v)
  )"

fun insert :: "int list \<Rightarrow> int \<Rightarrow> int list" where
  "insert nil v = cons v nil"
| "insert (cons v' vs) v = (
    if (v < v')
    then cons v (cons v' vs)
    else cons v' (insert vs v)
  )"

lemma tree_insert_list : "
  is_search_tree t \<Longrightarrow>
  insert (tree_to_list t) v = tree_to_list (tree_insert t v)
"
  sorry

end
